package net.knarcraft.permissionsigns.listener;

import net.knarcraft.permissionsigns.formatting.StringFormatter;
import net.knarcraft.permissionsigns.formatting.TranslatableMessage;
import net.knarcraft.permissionsigns.manager.SignManager;
import org.bukkit.Material;
import org.bukkit.Tag;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.block.Sign;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;

import java.util.ArrayList;
import java.util.List;

/**
 * A listener for relevant block events such as signs being broken
 */
public class BlockListener implements Listener {

    @EventHandler
    public void onBlockBreak(BlockBreakEvent event) {
        Block block = event.getBlock();
        Player player = event.getPlayer();

        if (event.isCancelled()) {
            return;
        }

        //if (block instanceof FallingBlock) {
        //TODO: Search recursively upwards (and downwards if pointed dripstone) until a non-falling block is found.
        // If upwards, check the non-falling block, but not downwards. If player does not have the create permission,
        // just stop and cancel once a permission sign is found. If it has, all affected signs need to be 
        // de-registered. Must account for stacks of sand and signs on top of each-other. So if a normal sign is 
        // found, the recursion upward must also happen.
        //}

        Material material = block.getBlockData().getMaterial();
        if (Tag.SIGNS.isTagged(material) || Tag.WALL_SIGNS.isTagged(material)) {
            checkIfBlockIsPermissionSign(block, player, event);
            if (event.isCancelled()) {
                return;
            }
        }

        Block relativeBlock = block.getRelative(BlockFace.UP);
        if (Tag.SIGNS.isTagged(relativeBlock.getBlockData().getMaterial())) {
            checkIfBlockIsPermissionSign(relativeBlock, player, event);
            if (event.isCancelled()) {
                return;
            }
        }

        if (block.getBlockData().getMaterial() == Material.POINTED_DRIPSTONE) {
            relativeBlock = block.getRelative(BlockFace.DOWN);
            if (Tag.WALL_SIGNS.isTagged(relativeBlock.getBlockData().getMaterial())) {
                checkIfBlockIsPermissionSign(relativeBlock, player, event);
                if (event.isCancelled()) {
                    return;
                }
            }
        }
        for (BlockFace blockFace : getRelevantBlockFaces()) {
            relativeBlock = block.getRelative(blockFace);
            if (Tag.WALL_SIGNS.isTagged(relativeBlock.getBlockData().getMaterial())) {
                checkIfBlockIsPermissionSign(relativeBlock, player, event);
                if (event.isCancelled()) {
                    return;
                }
            }
        }

        //TODO: Need to protect against other things that might damage the sign, such as explosions
    }

    /**
     * Check if the given block is a permission sign, and perform necessary tasks if it is
     *
     * @param block  <p>The block to check</p>
     * @param player <p>The player that caused the event</p>
     * @param event  <p>The triggered block break event</p>
     */
    private void checkIfBlockIsPermissionSign(Block block, Player player, BlockBreakEvent event) {
        Sign sign = (Sign) block.getState();
        boolean registered = SignManager.getSign(sign.getLocation()) != null;
        if (!registered) {
            return;
        }
        if (!player.hasPermission("permissionsigns.admin.create")) {
            event.setCancelled(true);
            player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.PERMISSION_SIGN_DESTROY_DENY));
        } else {
            if (!player.isSneaking()) {
                event.setCancelled(true);
                player.sendMessage(StringFormatter.getTranslatedErrorMessage(TranslatableMessage.PERMISSION_SIGN_DESTROY_SNEAK));
                return;
            }
            SignManager.removeSign(sign.getLocation());
            player.sendMessage(StringFormatter.getTranslatedInfoMessage(TranslatableMessage.PERMISSION_SIGN_REMOVED));
        }
    }

    /**
     * Gets the relevant block faces for checking around a block
     *
     * @return <p>The relevant block faces for checking around a block</p>
     */
    private List<BlockFace> getRelevantBlockFaces() {
        List<BlockFace> relevantBlockFaces = new ArrayList<>();
        relevantBlockFaces.add(BlockFace.WEST);
        relevantBlockFaces.add(BlockFace.EAST);
        relevantBlockFaces.add(BlockFace.NORTH);
        relevantBlockFaces.add(BlockFace.SOUTH);
        return relevantBlockFaces;
    }

}
